/* add.c - giis backend add function */
/* $OpenLDAP: pkg/ldap/servers/slapd/back-giis/add.c,v 1.6.2.3 2000/06/13 17:57:37 kurt Exp $ */
/*
 * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
 * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
 */

#include "portable.h"

#include <stdio.h>

#include <ac/string.h>
#include <ac/ctype.h>
#include <ac/socket.h>

#include <ldap_pvt.h>

#include "slap.h"
#include "slapd_giis.h"

static Filter   *giisstr2filter( const char *s );
static char     *find_matching_paren( const char *s );
static Filter   *str2list( const char *str, long unsigned int ftype);
static Filter   *str2simple( const char *str);
static int      str2subvals( const char *val, Filter *f);

static void     giisfilter_free( Filter *f);

int
giis_back_add(
    Backend	*be,
    Connection	*conn,
    Operation	*op,
    Entry	*e
)
{
	int inx, res, len, addflag=0;
	char *obj;
	Filter	*filter;
	Entry	*ent;
	GiisDT *gdt = (GiisDT *) be->be_private;
	int	total_rpolicy = gdt->total_rpolicy;
	char	**rpolicy = *gdt->rpolicy;

	if ( be->be_suffix ==NULL || be->be_suffix[0] == NULL) {
		send_ldap_result( conn, op, LDAP_UNWILLING_TO_PERFORM, NULL,
		    "suffix for the backend not found", NULL, NULL );
		return( -1 );
	}

	for (inx=0; inx<total_rpolicy; inx++) {
		filter = giisstr2filter(rpolicy[inx]);
		if (filter == NULL) {
			continue;
		}
		if ( test_filter( be, conn, op, e, filter ) == LDAP_COMPARE_TRUE ) {
			addflag = 1;
			if( filter ) filter_free( filter );
			break;
		}
		else if( filter ) {filter_free( filter ); filter=NULL;}
	}

	if (!addflag) {
		Debug( LDAP_DEBUG_ANY, "not authorized to add; check the policy file\n", 0, 0, 0 );
		send_ldap_result( conn, op, LDAP_UNWILLING_TO_PERFORM, NULL, "not authorized to add; check the policy file", NULL, NULL );
		return LDAP_UNWILLING_TO_PERFORM;
	}

        ldap_pvt_thread_mutex_lock( &entry2str_mutex );
	obj = entry2str( e, &len );
        ldap_pvt_thread_mutex_unlock( &entry2str_mutex );
	res = add_obj(gdt->glist, obj, &gdt->total_inx, &gdt->glist_mutex, gdt->giisgiven_sfx, gdt->local_port);

	if (res > 0) {
		send_ldap_result( conn, op, LDAP_SUCCESS, NULL, NULL, NULL, NULL );
		Debug( LDAP_DEBUG_ANY, "REGISTRATION SUCCEEDED\n", 0, 0, 0 );
		return LDAP_SUCCESS;
	}
	else {
		send_ldap_result( conn, op, LDAP_UNWILLING_TO_PERFORM, NULL, "registration failed", NULL, NULL );
		Debug( LDAP_DEBUG_ANY, "REGISTRATION FAILED\n", 0, 0, 0 );
		return LDAP_UNWILLING_TO_PERFORM;
	}
}


static Filter *
giisstr2filter( const char *str )
{
	Filter	*f = NULL;
	char	*end, *freeme;

	Debug( LDAP_DEBUG_FILTER, "giisstr2filter \"%s\"\n", str, 0, 0 );

	if ( str == NULL || *str == '\0' ) {
		return( NULL );
	}

	str = freeme = ch_strdup( str );

	switch ( *str ) {
	case '(':
		if ( (end = find_matching_paren( str )) == NULL ) {
			filter_free( f );
			free( freeme );
			return( NULL );
		}
		*end = '\0';

		str++;
		switch ( *str ) {
		case '&':
			Debug( LDAP_DEBUG_FILTER, "giisstr2filter: AND\n",
			    0, 0, 0 );

			str++;
			f = str2list( str, LDAP_FILTER_AND );
			break;

		case '|':
			Debug( LDAP_DEBUG_FILTER, "put_filter: OR\n",
			    0, 0, 0 );

			str++;
			f = str2list( str, LDAP_FILTER_OR );
			break;

		case '!':
			Debug( LDAP_DEBUG_FILTER, "put_filter: NOT\n",
			    0, 0, 0 );

			str++;
			f = str2list( str, LDAP_FILTER_NOT );
			break;

		default:
			Debug( LDAP_DEBUG_FILTER, "giisstr2filter: simple\n",
			    0, 0, 0 );

			f = str2simple( str );
			break;
		}
		*end = ')';
		break;

	default:	/* assume it's a simple type=value filter */
		Debug( LDAP_DEBUG_FILTER, "giisstr2filter: default\n", 0, 0,
		    0 );

		f = str2simple( str );
		break;
	}

	free( freeme );
	return( f );
}

/*
 * Put a list of filters like this "(filter1)(filter2)..."
 */

static Filter *
str2list( const char *str, unsigned long ftype )
{
	Filter	*f;
	Filter	**fp;
	char	*next;
	char	save;

	Debug( LDAP_DEBUG_FILTER, "str2list \"%s\"\n", str, 0, 0 );

	f = (Filter *) ch_calloc( 1, sizeof(Filter) );
	f->f_choice = ftype;
	fp = &f->f_list;

	while ( *str ) {
		while ( *str && isspace( (unsigned char) *str ) )
			str++;
		if ( *str == '\0' )
			break;

		if ( (next = find_matching_paren( str )) == NULL ) {
			filter_free( f );
			return( NULL );
		}
		save = *++next;
		*next = '\0';

		/* now we have "(filter)" with str pointing to it */
		if ( (*fp = giisstr2filter( str )) == NULL ) {
			filter_free( f );
			*next = save;
			return( NULL );
		}
		*next = save;

		str = next;
		fp = &(*fp)->f_next;
	}
	*fp = NULL;

	return( f );
}

static Filter *
str2simple( const char *str )
{
	Filter		*f;
	char		*s;
	char		*value, savechar;
	int			rc;
	const char		*text;

	Debug( LDAP_DEBUG_FILTER, "str2simple \"%s\"\n", str, 0, 0 );

	if ( (s = strchr( str, '=' )) == NULL ) {
		return( NULL );
	}
	value = &s[1];

	*s-- = '\0';	/* we shouldn't be mucking with str */
	savechar = *s;

	f = (Filter *) ch_calloc( 1, sizeof(Filter) );

	switch ( *s ) {
	case '<':
		f->f_choice = LDAP_FILTER_LE;
		*s = '\0';
		break;
	case '>':
		f->f_choice = LDAP_FILTER_GE;
		*s = '\0';
		break;
	case '~':
		f->f_choice = LDAP_FILTER_APPROX;
		*s = '\0';
		break;
	case ':':
		f->f_choice = LDAP_FILTER_EXT;
		*s = '\0';
		return NULL;
		break;

	default:
		if ( ldap_pvt_find_wildcard( value ) == NULL ) {
			f->f_choice = LDAP_FILTER_EQUALITY;
		} else if ( strcmp( value, "*" ) == 0 ) {
			f->f_choice = LDAP_FILTER_PRESENT;
		} else {
			f->f_choice = LDAP_FILTER_SUBSTRINGS;
			f->f_sub = ch_calloc( 1, sizeof( SubstringsAssertion ) );
			rc = slap_str2ad( str, &f->f_sub_desc, &text );
			if( rc != LDAP_SUCCESS ) {
				filter_free( f );
				*(value-1) = '=';
				return NULL;
			}
			if ( str2subvals( value, f ) != 0 ) {
				filter_free( f );
				*(value-1) = '=';
				return( NULL );
			}
			*(value-1) = '=';
			return( f );
		}
		break;
	}

	if ( f->f_choice == LDAP_FILTER_PRESENT ) {
		rc = slap_str2ad( str, &f->f_desc, &text );
		if( rc != LDAP_SUCCESS ) {
			filter_free( f );
			*(value-1) = '=';
			return NULL;
		}
	} else {
		char *tmp;

		f->f_ava = ch_calloc( 1, sizeof( AttributeAssertion ) );
		f->f_av_desc = NULL;
		rc = slap_str2ad( str, &f->f_av_desc, &text );
		if( rc != LDAP_SUCCESS ) {
			filter_free( f );
			*(value-1) = '=';
			return NULL;
		}

		tmp = ch_strdup( value );
		ldap_pvt_filter_value_unescape( tmp );
		f->f_av_value = ber_bvstr( tmp );
	}

	*s = savechar;
	*(value-1) = '=';

	return( f );
}

static int
str2subvals( const char *in, Filter *f )
{
	char	*nextstar, *val, *freeme;
	int	gotstar;

	Debug( LDAP_DEBUG_FILTER, "str2subvals \"%s\"\n", in, 0, 0 );

	if( in == NULL ) return 0;

	val = freeme = ch_strdup( in );
	gotstar = 0;

	while ( val && *val ) {
		if ( (nextstar = ldap_pvt_find_wildcard( val )) != NULL )
			*nextstar++ = '\0';

		ldap_pvt_filter_value_unescape( val );

		if ( gotstar == 0 ) {
			if ( val && strlen(val) > 0 )
				f->f_sub_initial = ber_bvstrdup( val );
			else
				f->f_sub_initial = NULL;

		} else if ( nextstar == NULL ) {
			if ( val && strlen(val) > 0 )
				f->f_sub_final = ber_bvstrdup( val );
			else
				f->f_sub_final = NULL;

		} else {
			if ( val && strlen(val) > 0 )
				charray_add( (char ***) &f->f_sub_any, (char *) ber_bvstrdup( val ) );
		}

		gotstar = 1;
		val = nextstar;
	}

	free( freeme );
	return( 0 );
}

/*
 * find_matching_paren - return a pointer to the right paren in s matching
 * the left paren to which *s currently points
 */

static char *
find_matching_paren( const char *s )
{
	int	balance, escape;

	balance = 0;
	escape = 0;
	for ( ; *s; s++ ) {
		if ( escape == 0 ) {
			if ( *s == '(' )
				balance++;
			else if ( *s == ')' )
				balance--;
		}
		if ( balance == 0 ) {
			return (char *) s;
		}
		if ( *s == '\\' && ! escape )
			escape = 1;
		else
			escape = 0;
	}

	return NULL;
}

/* should look at this for memory leak & crash */
static void
giisfilter_free( Filter *f )
{
        Filter  *p, *next;

        if ( f == NULL ) {
                return;
        }

        switch ( f->f_choice ) {
        case LDAP_FILTER_PRESENT:
                ad_free( f->f_desc, 1 );
                break;

        case LDAP_FILTER_EQUALITY:
        case LDAP_FILTER_GE:
        case LDAP_FILTER_LE:
        case LDAP_FILTER_APPROX:
/*
                ava_free( f->f_ava, 1 );
*/
                break;

        case LDAP_FILTER_SUBSTRINGS:
                ad_free( f->f_sub_desc, 1 );
                if ( f->f_sub_initial != NULL ) {
                        ber_bvfree( f->f_sub_initial );
                }
                ber_bvecfree( f->f_sub_any );
/*
                if ( f->f_sub_final != NULL ) {
                        ber_bvfree( f->f_sub_final );
                }
*/
                if(f->f_sub) ch_free( f->f_sub );
                break;

        case LDAP_FILTER_AND:
        case LDAP_FILTER_OR:
        case LDAP_FILTER_NOT:
                for ( p = f->f_list; p != NULL; p = next ) {
                        next = p->f_next;
                        giisfilter_free( p );
                }
                break;

        case SLAPD_FILTER_COMPUTED:
                break;

        default:
                Debug( LDAP_DEBUG_ANY, "giisfilter_free: unknown filter type=%lu\n",
                       f->f_choice, 0, 0 );
                break;
        }

        free( f );
}


char *
relable_dn(char *dn, char *sfx)
{
	int dn_size, sfx_size, inx, tinx, no, cnt;
	char *cdn, *res, *strinx;

	if (sfx == NULL || dn == NULL) {
		return NULL;
	}

	dn_size = strlen(dn);
	sfx_size = strlen(sfx);

	inx = 0;
	tinx = 0;
	while ((inx = get_position((dn+inx), "Mds-Vo-name=local")) > 0) {
		tinx = inx + tinx;
		inx = tinx;
	}
	if (tinx > 0) inx = tinx;

	if (inx <= -1) return NULL;
	if (inx > 0) {
		res = (char *) calloc(inx + sfx_size, sizeof(char));
		if (!res) {
			fprintf(stderr, "memory allocation error\n");
			Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
			return NULL;
		}
		for (no = 0; no < inx - 1; no++)
			res[no] = dn[no];
		cnt = 0;
		for ( ;no < inx + sfx_size - 1; no++)
			res[no] = sfx[cnt++];
		res[no] = 0;

		return res;
	}
	else {
		inx = 0;
		tinx = 0;
		while ((inx = get_position((dn+inx), "o=Grid")) > 0) {
			tinx = inx + tinx;
			inx = tinx;
		}
		if (tinx > 0) inx = tinx;
		if (inx <= 0) {
			return NULL;
		}
		else {
			res = (char *) calloc(inx + sfx_size, sizeof(char));
			if (!res) {
				fprintf(stderr, "memory allocation error\n");
				Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
				return NULL;
			}
			for (no = 0; no < inx - 1; no++)
				res[no] = dn[no];
			cnt = 0;
			for (;no < inx + sfx_size - 1; no++)
				res[no] = sfx[cnt++];
			res[no] = 0;

			return res;
		}
	}
}


int get_position(char *dn, char *pattern)
{

	int i, size;
	char *tdn, *tpattern, *strinx;

	if (!dn) return -1;

	tdn = (char *) strdup(dn);
	if (!tdn) {
		fprintf(stderr, "memory allocation failed\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return -1;
	}

	size = strlen(tdn);
	for (i=0; i<size; i++)
		tdn[i] = toupper(tdn[i]);

	if (!pattern) return -1;

	tpattern = (char *) strdup(pattern);
	if (!tpattern) {
		fprintf(stderr, "memory allocation failed\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return -1;
	}

	size = strlen(tpattern);
	for (i=0; i<size; i++)
		tpattern[i] = toupper(tpattern[i]);

	strinx = strstr(tdn, tpattern);
	if (strinx != NULL) {
		if(tdn) free(tdn);
		if(tpattern) free(tpattern);
		if ((int) (strinx - tdn) < 0) return -1;
		return (int) (strinx - tdn + 1);
	}
	else {
		if(tdn) free(tdn);
		if(tpattern) free(tpattern);
		return 0;
	}
}


char *
relable_object(char *obj, char *sfx)
{
	char *tdn, *dn, *res;
	int inx, dn_size, back_size, tdn_size, back_pos;

	if (!obj) return NULL;

	tdn = get_attribute(obj, "dn:", NULL);
	if(!tdn) return NULL;

	inx = get_position(obj, tdn);
	if (inx <= 0) {
		free(tdn);
		return NULL;
	}

	tdn_size = strlen(tdn);
	if (tdn_size <=0) {
		free(tdn);
		return NULL;
	}

	back_pos = tdn_size + inx - 1;
	back_size = strlen(obj + back_pos);
	if (back_size <=0) {
		free(tdn);
		return NULL;
	}

	dn = relable_dn(tdn, sfx);
	if(!dn) {
		free(tdn);
		return NULL;
	}

	dn_size = strlen(dn);
	if (dn_size <=0) {
		free(tdn);
		free(dn);
		return NULL;
	}

	res = (char *) calloc(back_size + dn_size + 5, sizeof(char));
	if (!res) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		free(tdn);
		free(dn);
		return NULL;
	}

	sprintf(res, "dn: %s%s", dn, obj + back_pos);

	free(tdn);
	free(dn);

	return res;
}


