/*
 * Copyright (c) 2002 Compaq Computer Corporation.  All Rights Reserved.
 *
 * Permission to use, copy, and modify this software and its documentation
 * is hereby granted under the following terms and conditions.  The above
 * copyright notice, this permission notice, and the following disclaimer
 * must appear in all copies of the software, derivative works or modified
 * versions.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND COMPAQ COMPUTER CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL COMPAQ COMPUTER
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <stdio.h>

#include <sys/sysinfo.h>
#include <sys/table.h>

#include <machine/hal_sysinfo.h>

const char *vendor = "Compaq Computer Corporation";

const char *getcpuname(unsigned long, const char **minor);
int dotable(int, unsigned long *, int *, unsigned long *);

/*
 * Usage: cpuinfo-tru64
 *
 * Do some table(2) and getsysinfo(2) in order to acquire information
 * about the CPUs installed on this machine.  Return results as shell
 * variable assignments suitable for "eval".  Variables assigned are:
 *
 *	_cpu_<N>_vendor		CPU vendor ('vendor' string)
 *	_cpu_<N>_model		CPU model (EV6, EV7, etc.)
 *	_cpu_<N>_version	CPU version (chip "pass" number)
 *	_cpu_<N>_speed		CPU speed in MHz
 *	_cpu_<N>_l2cache_size	Size of L2 cache in MBs (if available)
 *	_cpu_count		Number of CPUs
 *	_cpu_free1		1 minute load average
 *	_cpu_free5		5 minute load average
 *	_cpu_free15		15 minute load average
 *
 * Where <N> starts at "0" and is incremented for each additional CPU.
 *
 * Always exits with 0 (success).
 */
int
main(int argc, char *argv[])
{
	struct cpu_info c;
	struct tbl_loadavg loadtab;
	unsigned long runmask;
	const char *cname, *mname;
	int ci, slot, actmhz;
	unsigned long cpu, actcpu, actcsz;

	/* Get "default" cpu type */
	cpu = 0;
	ci = 0;
	(void)getsysinfo(GSI_PROC_TYPE, (caddr_t)&cpu, sizeof(cpu), NULL, NULL);

	/* Read in the 'cpu_info' structure */
	if (getsysinfo(GSI_CPU_INFO, (caddr_t)&c, sizeof(c), NULL, NULL) > 0) {
		slot = 0;

		/* only report on running CPUs */
		for (runmask = c.cpus_running; runmask; runmask >>= 1, slot++)
			if (runmask & 0x1) {
				if (!dotable(slot, &actcpu, &actmhz, &actcsz)) {
					actmhz = c.mhz;
					actcpu = cpu;
					actcsz = 0;
				}
				cname = getcpuname(actcpu, &mname);
				printf("_cpu_%d_vendor=\"%s\"\n", ci, vendor);
				printf("_cpu_%d_model=\"%s\"\n", ci, cname);
				printf("_cpu_%d_version=\"%s\"\n", ci, mname);
				printf("_cpu_%d_speed=\"%d\"\n", ci, actmhz);
				if (actcsz != 0)
					printf("_cpu_%d_l2cache_size=\"%ld\"\n",
						ci, actcsz / 1024);
				ci++;
			}
		printf("_cpu_count=\"%d\"\n", ci);
	}

	/* Read load average, result is integer normalized across CPUs. */
	if (table(TBL_LOADAVG, 0, &loadtab, 1, sizeof loadtab) >= 0) {
		slot = 100 * ci -
			100 * loadtab.tl_avenrun.l[0] / loadtab.tl_lscale;
		printf("_cpu_free1=\"%d\"\n", slot > 0? slot: 0);
		slot = 100 * ci -
			100 * loadtab.tl_avenrun.l[1] / loadtab.tl_lscale;
		printf("_cpu_free5=\"%d\"\n", slot > 0? slot: 0);
		slot = 100 * ci -
			100 * loadtab.tl_avenrun.l[2] / loadtab.tl_lscale;
		printf("_cpu_free15=\"%d\"\n", slot > 0? slot: 0);
	}

	return 0;
}

/*
 * getcpuname(cpu) - return CPU name that corresponds to 'cpu'.
 */
const char *
getcpuname(unsigned long cpu, const char **minor)
{
	static char cpubuf[64];
	static char minorbuf[64];
	static char *cpus[] = {
		NULL, "EV3", "EV4 (21064)",
		NULL, "LCA", "EV5 (21164)",
		"EV4/5 (21064A)", "EV5/6 (21164A)", "EV6 (21264)",
		"PCA56 (21164PC)", "PCA57 (21164PC)", "EV6/7 (21264A)",
		"EV6/8-CB (21264C)", "EV6/8-AF (21264B)", "EV6/8-CX (21264D)",
		"EV7 (21364)", "EV79 (21364A)", "EV69 (21264/EV69A)"
	};

	sprintf(minorbuf, "%lu", cpu >> 32);
	*minor = minorbuf;

	cpu &= 0xffffffffL;
	if (cpu < 0 || cpu >= sizeof(cpus) || cpus[cpu] == NULL) {
		sprintf(cpubuf, "unknown (type %lu)", cpu);
		return cpubuf;
	}

	return cpus[cpu];
}

/*
 * dotable() - return cpu type, speed and cache size using table(2).
 *
 * Use table(TBL_PROCESSOR_INFO) to gather information about a specific
 * processor.  Older systems do not provide the data we are interested
 * in so use a heuristic to determine whether the result is useful.
 *
 * Return 1 on success, 0 on failure.
 */
int
dotable(int slot, unsigned long *cpup, int *mhzp, unsigned long *cszp)
{
#ifdef	TBL_PROCESSOR_INFO
	struct my_tbl_processor {
		long _junk1[4];
		struct timeval junk2[2];
		int mhz;
		int _junk3;
		unsigned long num_cache_bytes;
		unsigned long type;
		long _junk5[47];
	} proctab;

	/*
	 * Initialize 'proctab.type' to zero.  If it changes, then the
	 * the fields we are interested in are valid.  Otherwise, go
	 * with the defaults.
	 */
	proctab.type = 0;
	if (table(TBL_PROCESSOR_INFO, slot, &proctab, 1, sizeof proctab) < 0)
		return 0;
	if (proctab.type == 0)
		return 0;

	*cpup = proctab.type;
	*mhzp = proctab.mhz;
	*cszp = proctab.num_cache_bytes;
	return 1;
#else
	return 0;
#endif
}

