/*
 * Copyright (c) 2002 Hewlett-Packard Company.  All Rights Reserved.
 *
 * Permission to use, copy, and modify this software and its documentation
 * is hereby granted under the following terms and conditions.  The above
 * copyright notice, this permission notice, and the following disclaimer
 * must appear in all copies of the software, derivative works or modified
 * versions.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND HEWLETT-PACKARD CO. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL HEWLETT-PACKARD
 * COMPANY BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include <sys/param.h>
#include <sys/pstat.h>
#include <sys/utsname.h>

char *vendor = "Hewlett-Packard Company";

/* N.B. This program will compile in a non-AnsiC environment */

/*
 * getcpuinfo(model, version) - return CPU model/version/feature strings.
 */
void
getcpuinfo(cpuidx, mod_p, ver_p, feat_p)
	int cpuidx;
	char **mod_p, **ver_p, **feat_p;
{
	static int firstime = 1;
	static char model[128], version[128], feature[128];
	struct utsname uts;
	char *vbuf1, *vbuf2, *cp, line[256];
	FILE *fp;
	long confret;

	/* It would appear that HP platforms do not support mixed CPUs */
	if (firstime == 0)
		goto usecache;

	firstime = 0;
	*model = *version = *feature = '\0';

	/* Get model information */
	if (uname(&uts) >= 0) {
		if ((vbuf1 = strchr(uts.machine, '/')) == NULL) {
			strncpy(model, uts.machine, sizeof model);
		} else {
			vbuf1++;
			if ((fp = fopen("/opt/langtools/lib/sched.models",
						"r")) != NULL ||
			    (fp = fopen("/usr/lib/sched.models",
						"r")) != NULL) {
				while (fgets(line, sizeof line, fp) != NULL) {
					cp = line;
					if (*cp==';' || *cp=='#' || *cp=='\n')
						continue;

#define	GETFIELD(_cp)							\
					while (*_cp != '\0' && !isspace(*_cp)) \
						_cp++;			\
					if (*_cp == '\0')		\
						continue;		\
					*_cp++ = '\0'
#define	NXTFIELD(_cp)							\
					while (isspace(*_cp))		\
						_cp++

					GETFIELD(cp);
					if (strcmp(line, vbuf1) != 0)
						continue;
					NXTFIELD(cp);
					GETFIELD(cp);
					NXTFIELD(cp);
					vbuf2 = cp;
					GETFIELD(cp);
					strncpy(model, vbuf2, sizeof model);
					break;
#undef NXTFIELD
#undef GETFIELD
				}
				fclose(fp);
			}
			if (*model == '\0')	/* cant match this model */
				snprintf(model, sizeof model, "(%s)", vbuf1);
		}
	}

	/* Get CPU version information */
	if ((confret = sysconf(_SC_CPU_VERSION)) != -1) {
		vbuf1 = ((confret & CPU_PA_RISC_MAX) != confret)?
			"Intel": (confret < CPU_PA_RISC1_0)?
			"Motorola": "PA-RISC";
		switch(confret) {
		    case CPU_HP_MC68020:	vbuf2 = "MC68020"; break;
		    case CPU_HP_MC68030:	vbuf2 = "MC68030"; break;
		    case CPU_HP_MC68040:	vbuf2 = "MC68040"; break;
		    case CPU_PA_RISC1_0:	vbuf2 = "1.0"; break;
		    case CPU_PA_RISC1_1:	vbuf2 = "1.1"; break;
		    case CPU_PA_RISC1_2:	vbuf2 = "1.2"; break;
#ifdef CPU_PA_RISC2_0
		    case CPU_PA_RISC2_0:
#else
		    case 0x214:
#endif
						vbuf2 = "2.0"; break;
#ifdef CPU_HP_INTEL_EM_1_0
		    case CPU_HP_INTEL_EM_1_0:
#else
		    case 0x300:
#endif
						vbuf2 = "1.0"; break;
#ifdef CPU_HP_INTEL_EM_1_1	/* just guessing... */
		    case CPU_HP_INTEL_EM_1_1:	vbuf2 = "1.1"; break;
#endif
#ifdef CPU_HP_INTEL_EM_1_2	/* just guessing... */
		    case CPU_HP_INTEL_EM_1_2:	vbuf2 = "1.2"; break;
#endif
#ifdef CPU_HP_INTEL_EM_2_0	/* just guessing... */
		    case CPU_HP_INTEL_EM_2_0:	vbuf2 = "2.0"; break;
#endif
		    default:			vbuf2 = "?.?"; break;
		}
		sprintf(version, "%s %s", vbuf1, vbuf2);
	}

	/* Get any special processor features */
	if ((confret = sysconf(_SC_KERNEL_BITS)) != -1)
		sprintf(feature, "%ld-bit", confret);

	if ((confret = sysconf(_SC_CPU_KEYBITS1)) != -1) {
		if ((confret & HARITH) != 0)
			strcat(feature, " HARITH");
		if ((confret & HSHIFTADD) != 0)
			strcat(feature, " HSHIFTADD");
	}

usecache:
	*mod_p = model;
	*ver_p = version;
	*feat_p = feature;
}

/*
 * Usage: cpuinfo_hpux_helper
 *
 * Acquire information about the CPUs installed on this machine.
 * Return results as shell variable assignments suitable for "eval".
 * Variables assigned are:
 *
 *	_cpu_<N>_vendor		CPU vendor ('vendor' string)
 *	_cpu_<N>_model		CPU model
 *	_cpu_<N>_version	CPU version
 *	_cpu_<N>_speed		CPU speed in MHz
 *	_cpu_<N>_features	CPU features
 *	_cpu_count		Number of CPUs
 *	_cpu_free1		1 minute load average
 *	_cpu_free5		5 minute load average
 *	_cpu_free15		15 minute load average
 *
 * Where <N> starts at "0" and is incremented for each additional CPU.
 *
 * Always exits with 0 (success).
 */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct pst_processor pstproc;
	struct pst_dynamic pstdyn;
	char *model, *version, *features;
	long ci, i;

	ci = 1;

	/* Load averages normalized across CPU count */
	if (pstat_getdynamic(&pstdyn, sizeof(pstdyn), 1, 0) != -1) {
		ci = (int) pstdyn.psd_proc_cnt;

		i = 100 * ci - 100 * pstdyn.psd_avg_1_min;
		printf("_cpu_free1=\"%ld\"\n", i > 0? i: 0);
		i = 100 * ci - 100 * pstdyn.psd_avg_5_min;
		printf("_cpu_free5=\"%ld\"\n", i > 0? i: 0);
		i = 100 * ci - 100 * pstdyn.psd_avg_15_min;
		printf("_cpu_free15=\"%ld\"\n", i > 0? i: 0);
	}
	printf("_cpu_count=\"%ld\"\n", ci);

	/* Processor information */
	for (i = 0; i < ci; i++) {
		if (pstat_getprocessor(&pstproc, sizeof(pstproc), 1, i) != -1) {
			getcpuinfo(i, &model, &version, &features);
			printf("_cpu_%ld_vendor=\"%s\"\n", i, vendor);
			printf("_cpu_%ld_model=\"%s\"\n", i, model);
			printf("_cpu_%ld_version=\"%s\"\n", i, version);
			printf("_cpu_%ld_features=\"%s\"\n", i, features);
			printf("_cpu_%ld_speed=\"%ld\"\n", i,
				pstproc.psp_iticksperclktick / 10000L);
		}
	}

	return 0;
}
