/*
 * Copyright (c) 2002 Hewlett-Packard Company.  All Rights Reserved.
 *
 * Permission to use, copy, and modify this software and its documentation
 * is hereby granted under the following terms and conditions.  The above
 * copyright notice, this permission notice, and the following disclaimer
 * must appear in all copies of the software, derivative works or modified
 * versions.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND HEWLETT-PACKARD CO. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL HEWLETT-PACKARD
 * COMPANY BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <errno.h>

/*
 * usage: grisdaemon user command [arguments ...]
 *
 * Launch a command (with optional arguments) as the specified user.
 * Any open file descriptors are first closed and the program is
 * disassociated from the terminal and given a new process group.
 */
int
main(int argc, char *argv[], char *envp[])
{
	struct passwd *pw;
	char *progname;
	int fd, fdmax, saverrno;

	/* set progname to path-stripped command name */
	if ((progname = strrchr(*argv, '/')) != NULL)
		progname++;
	else
		progname = *argv;

	if (argc < 3) {
		fprintf(stderr, "usage: %s user command [arguments ...]\n",
			progname);
		return 1;
	}


	/* skip setuid if specified name is - */
	if(strcmp("-",argv[1]))
	{

		/* does the target user exist? */
		if ((pw = getpwnam(argv[1])) == NULL) {
			fprintf(stderr, "%s: getpwnam: no such user: %s\n",
			progname, argv[1]);
			return 1;
		}

		/* become the target user */
		initgroups(argv[1], pw->pw_gid);
#ifdef __alpha		/* some vendors (e.g., Sun) define BSD empty */
#if defined(BSD) && (BSD == 198911)	/* assume Compaq Tru64 (not Linux) */
		setregid(pw->pw_gid, -1);
		setreuid(pw->pw_uid, -1);
#else					/* assume rest of world */
		setgid(pw->pw_gid);
		setuid(pw->pw_uid);
#endif
#endif

	}

	/* start new session with us as process group leader */
	(void) setsid();

	(void) chdir("/");

	/* close any open file descriptors (including passwd opened above) */
	fdmax = getdtablesize();
	for (fd = 0; fd < fdmax; fd++)
		(void) close(fd);

	/* open stdin/stdout/stderr */
	if ((fd = open("/dev/null", O_RDWR, 0)) != -1) {
		(void) dup2(fd, 0);
		(void) dup2(fd, 1);
		(void) dup2(fd, 2);
		if (fd > 2)
			(void) close(fd);
	}

	/* launch the command */
	execvp(argv[2], &argv[2]);
	saverrno = errno;

	/* report failure (use syslog, since stderr has been closed) */
	openlog(progname, LOG_CONS, LOG_DAEMON);
	errno = saverrno;
	syslog(LOG_ERR, "execvp: %s: %m\n", argv[2]);
	closelog();
	return 1;
}
